"""Experiment analysis for DerivaML.

This module provides the Experiment class for analyzing completed executions.
"""

from deriva_ml.experiment.experiment import Experiment

__all__ = ["Experiment"]
